/***************************************************************************
 *   Copyright (C) 2007 by Pelladi Gabor   *
 *   pelladigabor@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <list>
#include <vector>
#include <assert.h>
#include <time.h>

#include "ai_impl.h"
#include "standing.h"

// external parameters
pos_T table_size = 25;
int start_depth = 6;
int max_depth = 6;
int depth_increment = 2;
unsigned int max_branch = 100;
bool force_thinking = false;

bool cancelling = false;

// application initialization indicator
bool inited = false;
// thie standing persisted across steps
Standing rememberedStanding;
// alpha-beta current search depth limit
index_T depthLimit;
// steps made in the current game
int stepCount;
// step suggestion in the current depth limit
pos_T suggestedX, suggestedY;
// global best step
pos_T bestX, bestY;
// heuristic value of the search tree root
heur_T rootValue;

// types and classes
typedef std::list<Standing*> steps_T;
typedef std::vector<Standing*> sorting_steps_T;
typedef index_T entry_type_T;
enum { exact = 1, lower_bound = 2, upper_bound = 3 };

typedef unsigned long long hash_T;

#pragma pack(push, 1)
struct NodeHashData {
	hash_T checksum;
	heur_T value;
	index_T remaining_depth;
	entry_type_T entry_type;
};
#pragma pack(pop)

// hash table
const hash_T nodeHashMem = 16 * 1024 * 1024;
const hash_T nodeHashSize = nodeHashMem / sizeof(NodeHashData);
NodeHashData hashData[nodeHashSize];

const int hashMaxDepth = 8;

// a node of the alphabeta tree
class Node {
public:
	Standing* standing;

	Node* parent;
	Node* child;
	index_T depth;

	heur_T signum;
	heur_T alpha;
	heur_T beta;
	bool is_exact;

	bool evaluated;

	// the good steps that can be done from here
	steps_T steps;

	// constructors
	Node(Standing* _standing); // root
	Node(Standing* _standing, Node* _parent); //child

	// destructor
	~Node();

	void generateSteps();
	void calcHash(hash_T* hash, NodeHashData* data);
};

Node::Node(Standing* _standing):
	standing(_standing),
	parent(NULL), child(NULL), depth(0),
	signum(_standing->current == 0 ? 1 : -1), alpha(MinHeur - 1), beta(MaxHeur + 1), is_exact(false),
	evaluated(false)
{
	generateSteps();
}

Node::Node(Standing* _standing, Node* _parent):
	standing(_standing),
	parent(_parent), child(NULL), depth(_parent->depth + 1),
	signum(-(_parent->signum)), alpha(_parent->alpha), beta(_parent->beta), is_exact(false),
	evaluated(false)
{
	generateSteps();
}

Node::~Node() {
	if (child) delete child;
	if (standing) delete standing;
	for (steps_T::iterator i = steps.begin(); i != steps.end(); i++) delete (*i);
}

void Node::generateSteps() {
	steps.clear();
	if (depth >= depthLimit || standing->target) {
		heur_T hv = standing->hval;
		heur_T decay = depth_decay * depth;
		if (hv > 0) hv -= decay; else
		if (hv < 0) hv += decay;
		if (-decay <= hv && hv <= decay) hv = 0;
		alpha = beta = hv;
		evaluated = true;
		return;
	}

	suggestions_T suggestions;
	standing->getSuggestions(suggestions);

	sorting_steps_T sorting_steps;
	for (suggestions_T::iterator s = suggestions.begin(); s != suggestions.end(); s++) {
		pos_T x = s->x;
		pos_T y = s->y;
		assert(!standing->table[x][y]);
		Standing* p = new Standing(this->standing);
		p->step(x, y);

		heur_T pv = p->hval * signum;
		sorting_steps_T::iterator i;
		for (i = sorting_steps.begin(); i != sorting_steps.end() && signum * (*i)->hval > pv; i++) ;
		sorting_steps.insert(i, p);

		// we only keep the best looking positions
		if (sorting_steps.size() > max_branch) {
			delete sorting_steps.back();
			sorting_steps.pop_back();
		}
	}
	assert(sorting_steps.size() > 0);

	for (int i = sorting_steps.size() - 1; i >= 0; i--)
		steps.push_front(sorting_steps[i]);
}

void Node::calcHash(hash_T* hash, NodeHashData* data) {
	data->remaining_depth = depthLimit - depth;
	assert(data->remaining_depth > 0);
	
	hash_T c_hash = 0;
	for (int i = 0; i < table_size; i++) {
		for (int j = 0; j < table_size; j++) {
			c_hash = c_hash * 3145 + standing->table[i][j];
		}
	}
	data->checksum = c_hash;
	
	c_hash = 0;
	for (int i = 0; i < table_size; i++) {
		for (int j = 0; j < table_size; j++) {
			c_hash = (c_hash * 5313 + standing->table[i][j]) % nodeHashSize;
		}
	}
	*hash = c_hash;
}

void ai_init() {
	if (!inited) {
		srand(time(NULL));
		memset(hashData, 0, sizeof(hashData));
		inited = true;
	}
}

void ai_end() {
}

// main functions
void ai_new_game() {
	ai_init();
	stepCount = 0;
	rememberedStanding = Standing();
}

void ai_step(pos_T x, pos_T y) {
	stepCount++;
	rememberedStanding.step(x, y);
}

void ai_step_server(pos_T x, pos_T y) {
	rememberedStanding.step_server(x, y);
}

Field ai_think() {
	cancelling = false;

	// opening book
	if (stepCount == 0) {
		pos_T x, y;
		x = y = table_size / 2;
		while (rememberedStanding.table[x][y]) x++;
		return Field(x, y);
	} else if (stepCount == 1) {
		pos_T x, y;
		x = rememberedStanding.lastx;
		y = rememberedStanding.lasty;
		if (x < table_size / 2) {
			x++;
		} else {
			x--;
		}
		if (y < table_size / 2) {
			y++;
		} else {
			y--;
		}
		while (rememberedStanding.table[x][y]) x++;
		assert(x < table_size);
		return Field(x, y);
	}

	// alpha-beta pruning
	Node* act;
	Node* root;
	hash_T currentHash;
	NodeHashData currentHashData;

	for (depthLimit = start_depth; depthLimit <= max_depth && !cancelling; depthLimit += depth_increment) {
		printf("  depth limit: %2d", depthLimit);

		act = root = new Node(new Standing(rememberedStanding));

		// this prevents the AI from thinking if there is only one good move
		// should be commented out if there is a time limit for each move
		if (!force_thinking && root->steps.size() == 1) {
			suggestedX = root->steps.front()->lastx;
			suggestedY = root->steps.front()->lasty;
			rootValue = 0;
			act = NULL;
		}

		while (act && !cancelling) {
			// if this is a parent whose child has just been evaluated
			if (act->child) {
				if (act->signum > 0) {
					// MAX
					if (act->alpha < act->child->beta) {
						act->alpha = act->child->beta;
						act->is_exact = true;
						if (!act->parent) {
							suggestedX = act->child->standing->lastx;
							suggestedY = act->child->standing->lasty;
							rootValue = act->alpha;
						}
						if (act->alpha >= act->beta || act->alpha >= WinTreshold) {
							act->evaluated = true;
							currentHashData.entry_type = lower_bound;
							currentHashData.value = act->alpha;
						}
					}
				} else {
					// MIN
					if (act->beta > act->child->alpha) {
						act->beta = act->child->alpha;
						act->is_exact = true;
						if (!act->parent) {
							suggestedX = act->child->standing->lastx;
							suggestedY = act->child->standing->lasty;
							rootValue = act->beta;
						}
						if (act->alpha >= act->beta || act->beta <= -WinTreshold) {
							act->evaluated = true;
							currentHashData.entry_type = upper_bound;
							currentHashData.value = act->beta;
						}
					}
				}

				delete act->child;
				act->child = NULL;
			}

			// if this parent has no more children to process
			if (act->steps.empty()) {
				act->evaluated = true;
				if (act->signum > 0) {
					// MAX
					currentHashData.entry_type = act->is_exact ? exact : upper_bound;
					currentHashData.value = act->alpha;
				} else {
					// MIN
					currentHashData.entry_type = act->is_exact ? exact : lower_bound;
					currentHashData.value = act->beta;
				}
			}

			if (act->evaluated) {
				// store the current standing in the hash table
				if (act->depth <= hashMaxDepth) {
					act->calcHash(&currentHash, &currentHashData);
					hashData[currentHash] = currentHashData;
				}

				act = act->parent;
			} else {
				act->child = new Node(act->steps.front(), act);
				act->steps.pop_front();
				act = act->child;

				// if this is a leaf
				if (act->evaluated) {
					act = act->parent;
					continue;
				}

				// check whether we have already evaluated the standing elsewhere
				if (act->depth <= hashMaxDepth) {
					act->calcHash(&currentHash, &currentHashData);
					NodeHashData* storedHashData = &hashData[currentHash];
					if (storedHashData->remaining_depth >= depthLimit - act->depth && currentHashData.checksum == storedHashData->checksum) {
						if (act->signum > 0) {
							// MAX
							if (storedHashData->entry_type == exact)
								act->alpha = act->beta = storedHashData->value;
							if (storedHashData->entry_type == lower_bound && act->alpha < storedHashData->value)
								act->alpha = storedHashData->value;
							if (storedHashData->entry_type == upper_bound && act->beta > storedHashData->value)
								act->beta = storedHashData->value;
						} else {
							// MIN
							if (storedHashData->entry_type == exact)
								act->alpha = act->beta = storedHashData->value;
							if (storedHashData->entry_type == upper_bound && act->beta > storedHashData->value)
								act->beta = storedHashData->value;
							if (storedHashData->entry_type == lower_bound && act->alpha < storedHashData->value)
								act->alpha = storedHashData->value;
						}
						if (act->alpha >= act->beta) {
							act->evaluated = true;
							act = act->parent;
							continue;
						}
					}
				}
			}
		}

		// check why we have left the while loop
		if (!act) {
			bestX = suggestedX;
			bestY = suggestedY;
			printf("    suggestion: (%2d, %2d)    score: %6d\n", suggestedX, suggestedY, rootValue);
			fflush(stdout);
		} else {
			printf("    time is over\n");
		}
		delete root;
	}
	
	assert(cancelling || !rememberedStanding.table[bestX][bestY]);
	return Field(bestX, bestY);
}
